﻿// ===================================================================================
// Copyright � Genetec Inc., 2020
// All rights reserved.
// May only be used in accordance with a valid Source Code License Agreement.
//
// Genetec reserves the right to refuse to provide technical support in the event the 
// macro has been modified in any way, and for any reason.
// ===================================================================================
// ===================================================================================
// Synergis Periodic Access Solution Macro
//      Version 1.1
// Compatible with Genetec Security Center 5.8 and above
//
// This macro will restrict the entrance of the direct members of the configured 
// cardholder group to the configured area to only once every X minutes. The cardholders 
// can still exit the area once they enter. The access rules created by the macro will
// be added under the configured partition.
// ===================================================================================

using System;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using System.Xml;

using Genetec.Sdk;
using Genetec.Sdk.AccessControl.AccessRules;
using Genetec.Sdk.Entities;
using Genetec.Sdk.Enumerations;
using Genetec.Sdk.Events.AccessPoint;
using Genetec.Sdk.Queries;
using Genetec.Sdk.Scripting;

public sealed class PeriodicAccessSolutionMacro : UserMacro
{
    public Guid RestrictedCardholderGroup { get; set; }

    public Guid RestrictedArea { get; set; }

    public Guid AccessRulesPartition { get; set; }

    public int DenyDurationInMinutes { get; set; }

    private TimeSpan DenyDuration => new TimeSpan(0, DenyDurationInMinutes, 0);

    private Guid AllCarholders => Guid.Parse("00000000-0000-0000-0000-000000000008");

    private Guid RootPartition => Guid.Parse("00000000-0000-0000-0000-00000000000b");

    /// <summary>
    /// Entry point of the macro. Provide an implementation of this method.
    /// </summary>
    public override void Execute()
    {
        MacroLogger.TraceInformation("Prefetching entities");
        LoadEntities(EntityType.Cardholder, true);
        LoadEntities(EntityType.Area, true);

        MacroLogger.TraceInformation("Initializing proxy event registration");
        Sdk.EventReceived += new EventHandler<EventReceivedEventArgs>(OnEventReceived);

        MacroLogger.TraceInformation("Macro is ready to process events");
        Thread.Sleep(Timeout.Infinite);
    }

    /// <summary>
    /// Called when the macro needs to clean up. 
    /// </summary>    
    protected override void CleanUp()
    {
        // Release objects created by the Execute method, unhook events, and so on.
        MacroLogger.TraceInformation("Cleaning up macro");
        Sdk.EventReceived -= new EventHandler<EventReceivedEventArgs>(OnEventReceived);
    }

    private void LoadEntities(EntityType entityType, bool downloadAllRelatedData)
    {
        Stopwatch watch = Stopwatch.StartNew();
        var entityTypeString = entityType.ToString();

        try
        {
            var doorQuery = Sdk.ReportManager.CreateReportQuery(ReportType.EntityConfiguration) as EntityConfigurationQuery;
            if (doorQuery != null)
            {
                MacroLogger.TraceDebug(string.Format("Query for {0} prepared", entityTypeString));
                doorQuery.EntityTypeFilter.Add(entityType);
                doorQuery.DownloadAllRelatedData = downloadAllRelatedData;
                MacroLogger.TraceDebug(string.Format("Starting query for {0} with DownloadAllRelatedData {1} - Adding them to cache", entityTypeString, downloadAllRelatedData));
                var results = doorQuery.Query();
            }
            else
            {
                MacroLogger.TraceWarning(string.Format("Query was null, could not prefetch {0}.", entityTypeString));
            }
        }
        catch (Exception ex)
        {
            MacroLogger.TraceError(string.Format("Error in LoadEntities, could not prefetch {0}.", entityTypeString), ex);
        }
        finally
        {
            MacroLogger.TracePerformance(string.Format("LoadEntities: Completed loading of {0}. Time elapsed {1}", entityTypeString, watch.Elapsed));
            watch.Stop();
        }
    }

    private void OnEventReceived(object sender, EventReceivedEventArgs e)
    {
        if (IsDoorEntryDetected(e))
        {
            var entryDetectedEvent = e.Event as DoorEntryDetectedEvent;
            ProcessAccess(entryDetectedEvent.Cardholder, entryDetectedEvent.SourceEntity);
        }
        else if (IsDoorEntryAssumed(e))
        {
            var entryAssumedEvent = e.Event as DoorEntryAssumedEvent;
            ProcessAccess(entryAssumedEvent.Cardholder, entryAssumedEvent.SourceEntity);
        }
    }

    private bool IsDoorEntryDetected(EventReceivedEventArgs eventReceivedEventArgs)
    {
        return eventReceivedEventArgs.EventType == EventType.DoorEntryDetected;
    }

    private bool IsDoorEntryAssumed(EventReceivedEventArgs eventReceivedEventArgs)
    {
        return eventReceivedEventArgs.EventType == EventType.DoorEntryAssumed;
    }

    private void ProcessAccess(Guid cardholderGuid, Guid sourceGuid)
    {
        MacroLogger.TraceInformation(string.Format("Processing access for cardholder {0} on source entity {1}", cardholderGuid, sourceGuid));
        var area = GetInAreaFromSource(sourceGuid);
        if (area == null || area.Guid != RestrictedArea)
        {
            return;
        }

        var cardholder = Sdk.GetEntity(cardholderGuid) as Cardholder;
        if (cardholder == null)
        {
            MacroLogger.TraceWarning(string.Format("Couldn't fetch cardholder {0}", cardholderGuid));
            return;
        }

        var cardholderGroups = cardholder.Groups.ToList();

        if (RestrictedCardholderGroup == AllCarholders || cardholderGroups.Contains(RestrictedCardholderGroup))
        {
            MacroLogger.TraceInformation(string.Format("Denying cardholder {0} from entering area {1} for the next {2} ", cardholder.Name, area.Name, DenyDuration));
            Sdk.QueueUpdateAndWait(() =>
            {
                try
                {
                    var accessRuleLocalStartTime = DateTime.Now;
                    var accessRuleLocalEndTime = accessRuleLocalStartTime.Add(DenyDuration);

                    var accessRuleName = string.Format("Denying Rule {0}-{1}-{2}", accessRuleLocalStartTime.ToString("g"), cardholder.Name, area.Name);
                    var accessRule = Sdk.CreateAccessRule(accessRuleName, AccessRuleType.Temporary);
                    accessRule.IsDenyingRule = true;
                    accessRule.RunningState = State.Running;
                    accessRule.StartTime = accessRuleLocalStartTime.ToUniversalTime();
                    accessRule.EndTime = accessRuleLocalEndTime.ToUniversalTime();

                    accessRule.Members.Add(cardholderGuid);
                    area.AddAccessRule(accessRule.Guid, AccessRuleSide.In);

                    if (AccessRulesPartition != RootPartition)
                    {
                        var selectedPartition = Sdk.GetEntity(AccessRulesPartition) as Partition;
                        var rootPartition = Sdk.GetEntity(RootPartition) as Partition;
                        if (rootPartition != null && selectedPartition != null)
                        {
                            selectedPartition.AddMember(accessRule);
                            rootPartition.RemoveMember(accessRule);
                        }
                        else
                        {
                            MacroLogger.TraceWarning(string.Format("Couldn't fetch either the root partition or the configured partition {0}", AccessRulesPartition));
                        }
                    }
                }
                catch (Exception ex)
                {
                    MacroLogger.TraceError(string.Format("Could not create temporary denying rule and associate it to cardholder {0} and area {1}!", area.Name, cardholder.Name), ex);
                }
            });
        }
    }

    private Area GetInAreaFromSource(Guid sourceGuid)
    {
        var sourceEntity = Sdk.GetEntity(sourceGuid);

        var accessPoint = sourceEntity as AccessPoint;
        if (accessPoint == null)
        {
            MacroLogger.TraceWarning(string.Format("Couldn't fetch access point {0}", sourceGuid));
            return null;
        }

        // AccessPointSide.In is only used on Areas
        if (accessPoint.AccessPointSide == AccessPointSide.In)
        {
            MacroLogger.TraceDebug(string.Format("Access point {0} is an access point IN. Fetching associated area.", sourceGuid));
            var area = Sdk.GetEntity(accessPoint.AccessPointGroup) as Area;
            return area;
        }
        else
        {
            MacroLogger.TraceDebug(string.Format("Access point {0} is not an access point IN. Not processing further.", sourceGuid));
            return null;
        }
    }
}
